----------------------------------------------------------------
-- SRT-restoring reducer (srt_reducer.adb)
-- use "finite_field.ads"
-- 
-- Computes X mod M
----------------------------------------------------------------
with Gnat.Io; use Gnat.Io;
with finite_fields; use finite_fields;
procedure srt_reducer is

function csa_sum(a, b, c: in natural) return natural is
binary_a, binary_b, binary_c, binary_sum: bit_vector(0 .. n+1);
quotient, accumulator: natural;
begin
--natural to binary:
quotient := a; 
for i in 0 .. n+1 loop binary_a(i) := quotient mod 2; quotient := quotient/2; end loop;
quotient := b;
for i in 0 .. n+1 loop binary_b(i) := quotient mod 2; quotient := quotient/2; end loop;
quotient := c;
for i in 0 .. n+1 loop binary_c(i) := quotient mod 2; quotient := quotient/2; end loop;
--carry save addition, first component:
for i in 0 .. n+1 loop binary_sum(i) := (binary_a(i) + binary_b(i) + binary_c(i)) mod 2; end loop;
--binary to natural
accumulator := 0;
for i in 0 .. n+1 loop accumulator := accumulator + binary_sum(i)*(2**i); end loop;
return accumulator; 
end csa_sum;

function csa_carry(a, b, c: in natural) return natural is
binary_a, binary_b, binary_c, binary_carry: bit_vector(0 .. n+1);
quotient, accumulator: natural;
begin
--natural to binary:
quotient := a; 
for i in 0 .. n+1 loop binary_a(i) := quotient mod 2; quotient := quotient/2; end loop;
quotient := b;
for i in 0 .. n+1 loop binary_b(i) := quotient mod 2; quotient := quotient/2; end loop;
quotient := c;
for i in 0 .. n+1 loop binary_c(i) := quotient mod 2; quotient := quotient/2; end loop;
--carry save addition, second component:
binary_carry(0) := 0;
for i in 1 .. n+1 loop binary_carry(i) := (binary_a(i-1) + binary_b(i-1) + binary_c(i-1)) / 2; end loop;
--binary to natural
accumulator := 0;
for i in 0 .. n+1 loop accumulator := accumulator + binary_carry(i)*(2**i); end loop;
return accumulator; 
end csa_carry;

function quotient(ss, sc, y: in natural) return integer is
ss_high, sc_high, t: natural;
begin
ss_high := ss / (2**(n-1));   
sc_high := sc / (2**(n-1)); 
t := (ss_high + sc_high) mod 8;
if t <= 2 then return 1;
elsif t < 7 then return -1; 
else return 0;
end if;
end quotient;

m, y, ss, sc, rs, rc, minus_y, z: natural;
x: integer;
begin
loop
Put("m = "); Get(m);   
Put("x = "); Get(x);    
----------------------------------------------------------------------
y := m*(2**(n-k));
minus_y := (2**(n+2)) - y;
if x >= 0 then ss := x; sc := 0; else ss := (2**(n+2))+ x; sc := 0; end if;
for i in 0 .. n-k loop
   if quotient(ss, sc, y) = 1 then rs := csa_sum(ss, sc, minus_y); rc := csa_carry(ss, sc, minus_y); 
   elsif quotient(ss, sc, y) = 0 then rs := csa_sum(ss, sc, 0); rc := csa_carry(ss, sc, 0); 
   else rs := csa_sum(ss, sc, y); rc := csa_carry(ss, sc, y); 
   end if;
   ss := 2*rs; sc := 2*rc;
end loop;
z := ((rs + rc) mod (2**(n+1))) / (2**(n-k));
if z >= 2**k then z := (z + m) mod 2**k; end if;
---------------------------------------------------------------------------
Put(x); Put(" mod "); Put(m); Put(" = "); Put(z); New_Line; New_Line;
end loop;
end srt_reducer;





