----------------------------------------------------------------------------
-- mod p192 reducer (mod_p192_reducer.vhd)
--
-- reduce a 384 bits number modulo p192 (2**192-2**64-1)
--
-- two arquitectures are provided:
--               - fast_circuit
--               - area_circuit
----------------------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity mod_p192_reducer is
port(
  x: in std_logic_vector(383 downto 0);
  z: out std_logic_vector(191 downto 0)
);
end mod_p192_reducer;

architecture fast_circuit of mod_p192_reducer is
  signal xx1, xx2, xx3, xx4: std_logic_vector(191 downto 0);
  signal xx12, xx34: std_logic_vector(192 downto 0);
  signal s: std_logic_vector(193 downto 0);
  constant p192: std_logic_vector(191 downto 0):= (64 => '0', others => '1');
  constant minus_p: std_logic_vector(192 downto 0):= 
                       (0 => '1', 64 => '1', 192 => '1', others => '0');
  constant minus_2p: std_logic_vector(193 downto 0):= 
                       (1 => '1', 65 => '1', 193 => '1',others => '0');
  constant minus_3p: std_logic_vector(194 downto 0):= 
                       (0 => '1', 1 => '1', 64 => '1', 65 => '1', 
                        192 => '1', 194 => '1',others => '0');
  constant ZEROS: std_logic_vector(63 downto 0):= (others => '0');
  signal z1: std_logic_vector(192 downto 0);
  signal z2: std_logic_vector(193 downto 0);
  signal z3: std_logic_vector(194 downto 0);

begin
  xx1 <= x(383 downto 320) & x(383 downto 320) & x(383 downto 320);
  xx2 <= x(319 downto 256) & x(319 downto 256) & ZEROS;
  xx3 <= ZEROS & x(255 downto 192) & x(255 downto 192);
  xx4 <= x(191 downto 0);

  xx12 <= ('0' & xx1) + xx2;
  xx34 <= ('0' & xx3) + xx4;
  s <= ('0' & xx12 + xx34);

  z3 <= ('0' & s) + minus_3p;
  z2 <= s + minus_2p;  
  z1 <= s(192 downto 0) + minus_p;

  process(z1,z2,z3,s)
  begin
    if z3(194) = '0' then
       z <= z3(191 downto 0);
    elsif z2(193) = '0' then
       z <= z2(191 downto 0);     
    elsif z1(192) = '0' then
       z <= z1(191 downto 0);
    else
       z <= s(191 downto 0);  
    end if;
  end process;

end fast_circuit;

architecture area_circuit of mod_p192_reducer is
  signal xx1, xx2, xx3, xx4: std_logic_vector(191 downto 0);
  signal xx12, xx34, ss: std_logic_vector(192 downto 0);
  signal s: std_logic_vector(193 downto 0);
  constant p192: std_logic_vector(191 downto 0):= (64 => '0', others => '1');
  constant minus_p: std_logic_vector(192 downto 0):= 
                       (0 => '1', 64 => '1', 192 => '1', others => '0');
  constant ZEROS64: std_logic_vector(63 downto 0):= (others => '0');
  constant ZEROS62: std_logic_vector(61 downto 0):= (others => '0');
  signal zz: std_logic_vector(192 downto 0);

begin
  xx1 <= x(383 downto 320) & x(383 downto 320) & x(383 downto 320);
  xx2 <= x(319 downto 256) & x(319 downto 256) & ZEROS64;
  xx3 <= ZEROS64 & x(255 downto 192) & x(255 downto 192);
  xx4 <= x(191 downto 0);

  xx12 <= ('0' & xx1) + xx2;
  xx34 <= ('0' & xx3) + xx4;
  s <= ('0' & xx12 + xx34);

  ss <= ('0' & s(191 downto 0)) + (s(193 downto 192) & ZEROS62 & s(193 downto 192));
  zz <= ss + minus_p;
  
  with zz(192) select z <= ss(191 downto 0) when '1', zz(191 downto 0) when others;

end area_circuit;