----------------------------------------------------------------------------
-- Two Power K minus a (2**K - a) Reducer (tpkma_reducer.vhd)
--
-- Computes the remainder (x mod m) using m = 2**K-a algorithm
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package tpkma_reducer_parameters is
constant N: natural := 12;
constant K: natural := 5;
constant T:natural := 8; -- T >= K + log2(n-k+1)
--constant N: natural := 64;
--constant K: natural := 8;
--constant T:natural := 14; -- T = K + log2(n-k+1)
constant ZERO: std_logic_vector(T-K-1 downto 0) := (others => '0');
constant LONG_ZERO: std_logic_vector(N-T-1 downto 0) := (others => '0');
constant VERY_LONG_ZERO: std_logic_vector(N-K-1 downto 0) := (others => '0');
end tpkma_reducer_parameters;


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.tpkma_reducer_parameters.all;
entity tpkma_reducer is
port (
  x: in std_logic_vector (N-1 downto 0);
  m: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (K-1 downto 0);
  done: out std_logic
);
end tpkma_reducer;

architecture rtl of tpkma_reducer is
signal q: std_logic_vector(N-K-1 downto 0);
signal r, sum: std_logic_vector(t-1 downto 0);
signal not_m, a: std_logic_vector(K-1 downto 0);
signal r_minus_m: std_logic_vector(k downto 0);
signal product: std_logic_vector(N-1 downto 0);
signal load, reload, update, q_equal_zero: std_logic;
type states is range 0 to 5;
signal current_state: states;

begin
--not_gates: for i in 0 to K-1 generate not_m(i) <= not(m(i)); end generate;
--a <= not_m +1;
a <= not (m) + '1';
product <= q*a;
sum <= (zero&product(K-1 downto 0)) + r;

registers: process(clk)
begin
  if clk'event and clk = '1' then
    if load = '1' then 
      q <= x(N-1 downto K); 
      r <= zero & x(K-1 downto 0); 
    elsif reload = '1' then 
      q <= (long_zero & r(T-1 downto k)); 
      r <= zero & r(K-1 downto 0);
    elsif update = '1' then 
      q <= product(N-1 downto k); 
      r <= sum;
    end if;
  end if;
end process registers;

q_equal_zero <= '1' when q = very_long_zero else '0';

r_minus_m <= ('0'&r(K-1 downto 0)) + ('1'&a);
with r_minus_m(k) select z <= r_minus_m(K-1 downto 0) when '0', r(K-1 downto 0) when others;

control_unit: process(clk, reset, current_state, q_equal_zero)
begin
  case current_state is
    when 0 to 1 => load <= '0'; reload <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; reload <= '0'; update <= '0'; done <= '0';
    when 3 => load <= '0'; reload <= '0'; update <= '1'; done <= '0';
    when 4 => load <= '0'; reload <= '1'; update <= '0'; done <= '0';
    when 5 => load <= '0'; reload <= '0'; update <= '0'; done <= '0';
  end case;
  
  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '1' then current_state <= 1; end if;
      when 1 => if start = '0' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if q_equal_zero = '1' then current_state <= 4; end if;
      when 4 => current_state <= 5;
      when 5 => if q_equal_zero = '1' then current_state <= 0; else current_state <= 3; end if;
    end case;
  end if;
end process;

end rtl;

